//
// f d i s k . c
//
// minimized for Tiny Debian
//
// 2001/02/28 W.Nishida
//

/* to avoid parse error before `__extension__' in <linux/string.h> included by
 * <linux/genhd.h> (conflicts with <bits/string2.h> on sparc at least).
 */
#define __NO_STRING_INLINES

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <mntent.h>
#include <assert.h>
#include <syslog.h>
#include <linux/cdrom.h>
#include <linux/major.h>
#include <linux/genhd.h>

#include "fdisk.h"

/*
 * XXX: Some 2.2.x kernel headers (ie. libc6-dev) do
 * not contain this for some reason -- Ben
 */
#ifndef MAX_DISKNAME_LEN
#define MAX_DISKNAME_LEN 32
#endif

struct fdisk_disk	*fdisk_disks = NULL;
struct fdisk_partition	*fdisk_partitions;
struct fdisk_partition	*fdisk_partitions_by_type[FSTYPE_MAX];
char			*fstype_name[FSTYPE_MAX];

struct fdisk_partition	*mounted_partitions;
struct fdisk_partition	*swapon_partitions;

struct fdisk_disk	*last_disk;
struct fdisk_partition	*last_partition;
struct fdisk_partition	*last_partitions_by_type[FSTYPE_MAX];
struct fdisk_partition	*last_swapon_partition;

/* ptype2fstype[] translates partition type numbers to file system types */

static struct fstypes {
    int ptype;
    int fstype;
} ptype2fstype[] = {
    /* PC */
    { PTYPE_DOSEXT,		FSTYPE_EXTPART },
    { PTYPE_WEXT,		FSTYPE_EXTPART },
    { PTYPE_LSWAP,		FSTYPE_SWAP },
    { PTYPE_LINUX,		FSTYPE_DEFAULT },
    { PTYPE_LINUXEXT,		FSTYPE_EXTPART },
    { PTYPE_NFS,		FSTYPE_NFS },
};

int
fdisk_fstype_of(unsigned int ptype)
{
    struct fstypes *p;
    
    for( p = ptype2fstype+SIZE(ptype2fstype)-1; p >= ptype2fstype; p-- ) {
	if (ptype == p->ptype)
	return p->fstype;
    }
    return FSTYPE_UNKNOWN;
}

/* sys_types[] translates system ID number to description */

static struct systypes {
    unsigned int type;
    char *name;
} sys_types[] = {
    {0, "Empty"},
    {5, "DOS Extended"},                /* DOS 3.3+ extended partition */
    {0x0f, "Win95 Extended (LBA)"},
    {0x82, "Linux swap"},
    {0x83, "Linux native"},
    {0x85, "Linux extended"},
    {PTYPE_NFS, "NFS"},
};

char *
fdisk_sysname(int type) {
    struct systypes *s;

    for( s = sys_types+SIZE(sys_types)-1; s >= sys_types; s-- )
      if (s->type == type)
        return s->name;
    return "Unknown";
}

/* fstype_name[] translates filesystem number to description */

char *fstype_name[FSTYPE_MAX] = {
    "ext2",
    "swap",
    "reiser",
    "nfs",
    "extpart", /* FSTYPE_EXTPART is not a valid filesystem, so return "UNKNOWN" */
    "unknown" /* FSTYPE_UNKNOWN */
};

char *
fdisk_fstype_name_of(unsigned int ptype)
{
    int fstype = fdisk_fstype_of(ptype);
    return fstype_name[fstype];
}

static struct DEVICE_BLOCK {
    dev_t dev;	/* initial device number */
    int n_dev;	/* number of drives in this block */
    int step;	/* step in device number between drives */
} device_blocks[] = {
    /* all the r/w block devices known to the kernel */
    /* according to linux/Documentation/devices.txt  */
    {  3 << 8,   2, 64 },	/* hda      - hdb       */
    {  8 << 8,  16, 16 },	/* sda      - sdp       */
    { 22 << 8,   2, 64 },	/* hdc      - hdd       */
    {       0,   0,  0 }
};

static struct {
    char *name;
    int (*func)(char *device, int fd);
} parsers[] = {
    { "msdos", parse_msdos_partition },
};

static void fdisk_init() {
  int i;
  struct fdisk_disk *d1,*d2;
  struct fdisk_partition *p1,*p2;

  if (NULL != fdisk_disks) {
    d1 = fdisk_disks;
    while ( d1 ) {
      d2 = d1;
      d1 = d1->next;
      free (d2);
    }
    p1 = fdisk_partitions;
    while ( p1 ) {
      p2 = p1;
      p1 = p1->next;
      free (p2);
    }
  }

  for (i=0; i<FSTYPE_MAX; i++) {
    fdisk_partitions_by_type[i] = NULL;
    last_partitions_by_type[i] = NULL;
  }

  fdisk_disks = last_disk = NULL;
  fdisk_partitions = last_partition = NULL;
}

struct fdisk_disk *
fdisk_add_disk(const char *name, unsigned long size){
  struct fdisk_disk *disk;
  
  disk = (struct fdisk_disk *) malloc(sizeof(struct fdisk_disk));
  if (!disk) return NULL;
  
  disk->name = strdup(name);
  disk->size = size;
  disk->next = NULL;
  disk->partitions = NULL;
  disk->last_partition = NULL;
  disk->partition_format = NULL;

  if (NULL == fdisk_disks) {
    fdisk_disks = disk;
  }

  if (NULL != last_disk) {
    last_disk->next = disk;
  }
  last_disk = disk;

#ifdef DEBUG
  fprintf(stderr, "fdisk: added disk %s, size %ld\n", name, size );
#endif

  return(disk);
}

/* get_part_info: read and parse line in /proc/partitions format */
static int get_part_info(FILE *f, dev_t *dev, char *name)
{
    char line[64];
    unsigned int major, minor;

    while (fgets(line, sizeof line, f)) {
	if (sscanf(line, "%4u  %4u %*u %s", &major, &minor, name) == 3) {
	    *dev = MKDEV(major, minor);
	    return 1;
	}
    }
    return 0;
}

/* strdupcat: return concatenation of two strings in malloc-ed buffer */
static char *strdupcat(char *s1, char *s2)
{
    char *out = malloc(strlen(s1) + strlen(s2) + 1);
    if (out) {
	strcpy(out, s1);
	strcat(out, s2);
    }
    return out;
}

/* part_name: return partition device name from disk and partition number */
static char *part_name(struct fdisk_disk *disk, unsigned int part)
{
    FILE *f;
    char *pname = NULL, *dname = disk->name + 5;

    /* scan /proc/partitions for likely devices */
    f = fopen("/proc/partitions", "r");
    if (f) {
	char name[MAX_DISKNAME_LEN];
	dev_t dev, disk_dev = 0;

	while (get_part_info(f, &dev, name)) {
	    /* look for target partition or real disk containing it */
	    if (! disk_dev) {
		/* look for real disk containing target partition */
		if (strcmp(name, dname) == 0)
		    disk_dev = dev;
	    } else {
		/* look for partition with same number as target */
		if (disk_dev + part == dev) {
#ifdef DEBUG
		    fprintf(stderr, "fdisk: found partition %s: major=%d, minor=%d\n",
			   name, (int)MAJOR(dev), (int)MINOR(dev));
#endif
		    pname = strdupcat("/dev/", name);
		    return pname;
		}
	    }
	}
	fclose(f);
    }
    return NULL;
}


struct fdisk_partition *fdisk_add_partition(const char *name, int minor,
 unsigned int type, unsigned long size)
{
    int fstype;
    struct fdisk_partition *partition = malloc(sizeof(struct fdisk_partition));

    if (!partition)
        return NULL;

    partition->name = NULL;
    partition->mount_point= NULL;
    partition->type = type;
    partition->size = size;
    partition->in_use = 0;
    partition->disk = NULL;
    partition->next = NULL;
    partition->next_by_disk = NULL;
    partition->next_by_type = NULL;
    partition->next_in_use = NULL;

    if (minor == -1) {	/* Non local disk paritions */
	partition->name = strdup(name);
	assert(partition->name);
     }
    else {		/* Partitions on the local disks */
	partition->disk = fdisk_find_disk(name);
	assert(partition->disk);
	assert(minor >= 0);
	partition->name = part_name(partition->disk, (unsigned) minor);
	if (! partition->disk->partitions)
	    partition->disk->partitions = partition;
	if (partition->disk->last_partition)
	    partition->disk->last_partition->next_by_disk = partition;
	partition->disk->last_partition = partition;
     }
  
    if (! fdisk_partitions)
        fdisk_partitions = partition;
    if (last_partition)
        last_partition->next = partition;
    last_partition = partition;

    fstype = fdisk_fstype_of(type);
    if (! fdisk_partitions_by_type[fstype])
        fdisk_partitions_by_type[fstype] = partition;
    if (last_partitions_by_type[fstype])
        last_partitions_by_type[fstype]->next_by_type = partition;
    last_partitions_by_type[fstype] = partition;

#ifdef DEBUG
    fprintf(stderr, "fdisk: added partition %s, size %ld, type %08x=\"%s\"\n",
	   partition->name, size, type, fdisk_sysname(type) );
#endif
    return partition;
}

struct fdisk_disk *
fdisk_find_disk(const char *name) {
  struct fdisk_disk *d;

  d = fdisk_disks;

  while (d) {
    if (0 == strcmp(name, d->name)) {
      return (d);
    }
    d = d->next;
  }

  return(NULL);
}

struct fdisk_partition*
fdisk_find_partition_by_mntpoint(const char* mount_point)
{
  struct fdisk_partition* p;

  if (mount_point) {
    p = mounted_partitions;

    while (p) {
      if (0 == strcmp(mount_point, p->mount_point))
        return p;
      p = p->next_in_use;
    }
  }

  return (NULL);
}

struct fdisk_partition *
fdisk_find_partition_by_name(const char *name){
  struct fdisk_partition *p;

  if (name) {
    p = fdisk_partitions;

    while (p) {
      if (0 == strcmp(name, p->name))
        return(p);
      p = p->next;
    }
  }

  return(NULL);
}

struct fdisk_partition *
fdisk_find_partition_by_type(const int type){
  struct fdisk_partition *p;

  p = fdisk_partitions;

  while (p) {
    if (p->type == type)
      return(p);
    p = p->next;
  }

  return(NULL);
} 

static void mounted_reread() {
  FILE *mtabFile;
  struct mntent *mnt_ent;
  struct fdisk_partition *p; 
  int type;
  
  mounted_partitions = NULL;

  if (! (mtabFile = setmntent(MTAB, "r")))
    return ;

  while ((mnt_ent = getmntent(mtabFile))) {
#ifdef DEBUG
    fprintf(stderr,"fdisk: mounted partition: %s, on: %s, type: %s , options: %s ...\n", mnt_ent->mnt_fsname, mnt_ent->mnt_dir, mnt_ent->mnt_type, mnt_ent->mnt_opts);
#endif
    /* If mnt_ent is proc, devpts, or none ignore it */
    if ( (strcmp(mnt_ent->mnt_fsname,"proc") == 0) ||
	 (strcmp(mnt_ent->mnt_fsname,"devpts") == 0) ||
	 (strcmp(mnt_ent->mnt_fsname,"none") == 0) ) {
      continue;
    }
    /* If mnt_ent is our current root fs, ignore it */
    if (strcmp(mnt_ent->mnt_dir,"/") == 0) {
      continue;
    }
    /* If mnt_ent is a partition on the local disks add it to the
     * partitions list*/
    p = fdisk_find_partition_by_name(mnt_ent->mnt_fsname); 
    if (! p) {		/* Not on the local disk */
	for (type = 0; fstype_name[type]; type++)
	    if (! strcmp(mnt_ent->mnt_type, fstype_name[type])) break;
	if (type == FSTYPE_UNKNOWN) continue;
	switch (type) {
	    case FSTYPE_NFS:
		type = PTYPE_NFS;
		break;
	 }
	p = fdisk_add_partition(mnt_ent->mnt_fsname, -1, type, 0);
     }
    p->in_use=-1;
    p->mount_point=strdup(mnt_ent->mnt_dir);
    if (mounted_partitions) {
    /* reverse ordered, as this is the order in which they would be
     * un-mounted. */
        p->next_in_use=mounted_partitions;
    }
    mounted_partitions = p; 
  }
  endmntent(mtabFile);
}

struct swpent *
getswpent (FILE *stream, struct swpent *sp)
{
  char *head;
  char buffer[BUFSIZE];

  do
    {
      char *end_ptr;

      if (fgets (buffer, BUFSIZE, stream) == NULL)
        return NULL;

      end_ptr = strchr (buffer, '\n');
      if (end_ptr != NULL)      /* chop newline */
        *end_ptr = '\0';
      else
        {
          /* Not the whole line was read.  Do it now but forget it.  */
          char tmp[1024];
          while (fgets (tmp, sizeof tmp, stream) != NULL)
            if (strchr (tmp, '\n') != NULL)
              break;
        }
      head = buffer;
      /* skip empty lines and comment lines:  */
    } while (head[0] != '/');

  head += strcspn (head, " \t");
  *head++='\0';
  sp->name = strdup(buffer);
  head += strspn (head, " \t");
  sp->type = (strncmp(head, "file", 4) == 0) ? 1 : 0;
  head += strcspn (head, " \t");
  sscanf(head," %d %d %d ", &sp->size, &sp->used, &sp->prio);

  return sp;
}

static void swapon_reread() {
  FILE *swapsFile;
  struct swpent swap_ent;

  struct fdisk_partition *p;

  swapon_partitions = last_swapon_partition = NULL;

  if (! (swapsFile = setmntent(STAB, "r")))
    return ;

  while (getswpent(swapsFile, &swap_ent)) {
#ifdef DEBUG
    fprintf(stderr,"fdisk: swap partition: %s, type: %s , size: %d, used: %d, prio: %d ... ",
	swap_ent.name, (swap_ent.type==0)?"partition":"file", swap_ent.size,
	swap_ent.used, swap_ent.prio);
#endif
    /* If swap_ent is not a partition on the local disks, then ignore it */
    p = fdisk_find_partition_by_name(swap_ent.name);
    free(swap_ent.name);
    if (p) {
      p->in_use=-1;
      if (! swapon_partitions)
        swapon_partitions = p;
      else
        last_swapon_partition->next_in_use = p;
      last_swapon_partition = p;
    }
  }
  fclose(swapsFile);
}

/*
 * sseek: seek to specified sector - return 0 on failure
 *
 * For >4GB disks lseek needs a > 32bit arg, and we have to use llseek.
 * On the other hand, a 32 bit sector number is OK until 2TB.
 * The routines _llseek and sseek below are the only ones that
 * know about the loff_t type.
 *
 * sseek() uses the _llseek system call directly (if available), because of
 * some libc5/libc6 mismatch stuff.
 */

static _syscall5( int, _llseek, uint, fd, ulong, hi, ulong, lo,
		  loff_t *, res, uint, wh );

/* seek to a sector */
static int sseek(int fd, unsigned long s)
{
    loff_t in, out;

    in = (loff_t)s * 512;
    out = 1;
    if (_llseek (fd, in>>32, in & 0xffffffff, &out, SEEK_SET) != 0)
	return 0;
    if (in != out)
	return 0;
    return 1;
}

int sread(int fd, unsigned long sector, void *buffer)
{
    if (!sseek( fd, sector ))
	return 0;
    if (read( fd, buffer, 512 ) != 512)
	return 0;
    return 1;
}

static void
parse_partition(char *device, int fd)
{
    int i, rv;

    for( i = 0; i < SIZE(parsers); ++i ) {
	rv = parsers[i].func( device, fd );
	if (rv < 0)
	    return;
	if (rv > 0) {
	    struct fdisk_disk *disk;
	    if ((disk = fdisk_find_disk(device)))
		disk->partition_format = parsers[i].name;
	    return;
	}
    }
}


/* ide_is_disk: return true if IDE device is a disk drive */
static int ide_is_disk(char *device)
{
    char pname[32];
    int n;
    FILE *f;
    int result = 0;

    /* /dev/hd? => /proc/ide/hd?/media */
    device = strrchr(device, '/');
    assert(device);
    n = snprintf(pname, sizeof pname, "/proc/ide/%s/media", device + 1);
    assert(n > 0);

    /* open /proc/ide/hd?/media and look for ide_disk text */
    f = fopen(pname, "r");
    if (f) {
	char line[8];
	result = fgets(line, sizeof line, f) && strcmp(line, "disk\n") == 0;
	fclose(f);
    }
    return result;
}

/* is_disk: return true if device is a disk drive */
static int is_disk(char *device, int fd)
{
    struct stat stat_buf;

    /* reject anything un-stat-able, or not a block device */
    if (fstat(fd, &stat_buf) || ! S_ISBLK(stat_buf.st_mode))
	return 0;

    /* IDE devices */
    switch (MAJOR(stat_buf.st_rdev)) {
    case IDE0_MAJOR: case IDE1_MAJOR: case IDE2_MAJOR:
    case IDE3_MAJOR:
	return ide_is_disk(device);
    }

    /* SCSI disk driver will only accept disks and mag-opt drives */
    switch (MAJOR(stat_buf.st_rdev)) {
    case SCSI_DISK0_MAJOR:
	return 1;
    }

    return 0;
}

static void alarmed(int signo) { }

/* try_device: check specific device by name, add to disk/part lists if ok */
static void try_device(char *device)
{
    int fd;

    signal(SIGALRM, alarmed);
    alarm(30);
    fd = open(device, O_RDONLY);
    alarm(0);

    if (fd >= 0) {
        if (is_disk(device, fd)) {
	    unsigned int size;
	    if (ioctl(fd, BLKGETSIZE, &size))
		size = 0;   /* don't complain, size not really used yet */
	    size >>= 1;     /* convert from 512-byte sectors to kB */
	    fdisk_add_disk(device, size);

	    parse_partition(device, fd);
	} else {
	    /* Don't need to do anything for individual partitions because   */
	    /* parse_partition() above calls fdisk_add_partition() for every */
	    /* partition on the 'whole-disk' device.  We could eliminate a   */
	    /* huge amount of code and complexity if we knew the partition   */
	    /* type right now.						     */
	}
	close(fd);
    }
}


/* is_drive: return true if device is a 'whole disk', not just a partition */
static int is_drive(dev_t dev)
{
    struct DEVICE_BLOCK *d;

    /* search for device in each block of devices */
    for (d = device_blocks; d->dev && dev >= d->dev; d++)
	if (((dev - d->dev) % d->step) == 0)
	    return 1;

    return 0;
}


/* fdisk_reread: build up disk/partition lists from scratch */
void fdisk_reread(void)
{
    FILE *f;

    fdisk_init();

    /* scan /proc/partitions for likely device names */
    f = fopen("/proc/partitions", "r");
    if (f) {
	char name[MAX_DISKNAME_LEN];
	dev_t dev;

	while (get_part_info(f, &dev, name)) {
	    /* check device number against known 'whole-disk' devices */
	    if (is_drive(dev)) {
		static char path[] = "/dev/";
		char dev_name[sizeof path + sizeof name];

		/* try to build partition list for device */
		strcpy(dev_name, path);
		strcat(dev_name, name);
		try_device(dev_name);
	    }
	}
	fclose(f);
    }

    mounted_reread();
    swapon_reread();
}
