//
// f d i s k . h
//
// minimized for Tiny Debian
//
// 2001/02/28 W.Nishida
//

#ifndef _FDISK_H_
#define _FDISK_H_

/*
 * XXX: Some 2.2.x kernel headers (ie. libc6-dev) do
 * not contain this for some reason -- Ben
 */
#ifndef MAX_DISKNAME_LEN
#define MAX_DISKNAME_LEN 32
#endif

#define BUFSIZE		8192

#define MTAB		"/proc/mounts"
#define STAB		"/proc/swaps"

#define FSTYPE_EXT2	0
#define FSTYPE_SWAP	1
#define FSTYPE_REISER	2
#define FSTYPE_NFS	3
#define FSTYPE_EXTPART	4
#define FSTYPE_UNKNOWN	5
#define FSTYPE_MAX	6
#define FSTYPE_DEFAULT	FSTYPE_EXT2
//#define FSTYPE_DEFAULT	FSTYPE_REISER

#define SIZE(a) (sizeof(a)/sizeof(a[0]))

struct fdisk_partition;

struct fdisk_disk {
  char *name;
  unsigned size; /* not yet used */
  char *partition_format;

  struct fdisk_partition *partitions;
  struct fdisk_partition *last_partition;

  struct fdisk_disk *next;
};

struct fdisk_partition {
  char *name;
  char *mount_point;
  int type; /* type code, see PTYPE_* macros below */
  unsigned size; /* in kByte */
  int in_use;

  struct fdisk_disk *disk;

  struct fdisk_partition *next;
  struct fdisk_partition *next_by_disk;
  struct fdisk_partition *next_by_type;
  struct fdisk_partition *next_in_use;
};

// Global variables

struct fdisk_disk       *fdisk_disks;
struct fdisk_partition  *fdisk_partitions;
struct fdisk_partition  *fdisk_partitions_by_type[ FSTYPE_MAX ];
char                    *fstype_name[ FSTYPE_MAX ];
struct fdisk_partition  *mounted_partitions;
struct fdisk_partition  *swapon_partitions;

// Global functions

int    fdisk_fstype_of(unsigned int ptype);
char  *fdisk_sysname(int type);
char  *fdisk_fstype_name_of(unsigned int ptype);

struct fdisk_disk *fdisk_find_disk(const char *name);
struct fdisk_partition *fdisk_find_partition_by_mntpoint(const char *mntpoint);
struct fdisk_partition *fdisk_find_partition_by_name(const char *name);
struct fdisk_partition *fdisk_find_partition_by_type(const int type);
void   fdisk_reread();

struct fdisk_disk *fdisk_add_disk(const char *name, unsigned long size);
struct fdisk_partition *fdisk_add_partition(const char *name, int minor, unsigned int type, unsigned long size);
int sread(int fd, unsigned long sector, void *buffer);
int parse_msdos_partition(char *device, int fd);

/* Partition type numbers */
#define PTYPE_UNKNOWN		0x00000100
#define PTYPE_PREFIX_MSDOS	0x00000000
#define PTYPE_EMPTY		0
#define PTYPE_DOSEXT		0x05
#define PTYPE_WEXT		0x0f
#define PTYPE_LSWAP		0x82
#define PTYPE_LINUX		0x83
#define PTYPE_LINUXEXT		0x85
#define PTYPE_NFS		0x4e465300

/* Swap entry */
struct swpent
  {
    char *name;
    int type; /* 0 = partition, 1 = file */
    int size;
    int used;
    int prio;
  };

#endif
