//
// p a r t b l _ m s d o s . c
//
// minimized for Tiny Debian
//
// 2001/03/02 W.Nishida
//

/*
 * glibc 2.1 strxxx macros conflicts with linux/string.h included from
 * asm/unaligned.h if the following name is not defined
 */
#define __NO_STRING_INLINES

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <asm/types.h>
#include "byteorder.h"

#include "fdisk.h"

/*
 * Many architectures don't like unaligned accesses, which is
 * frequently the case with the nr_sects and start_sect partition
 * table entries.
 */
#include <asm/unaligned.h>

#define SYS_IND(p)	(get_unaligned(&p->sys_ind))
#define NR_SECTS(p)	({ __typeof__(p->nr_sects) __a =	\
				get_unaligned(&p->nr_sects);	\
				__le32_to_cpu(__a); \
			})

#define START_SECT(p)	({ __typeof__(p->start_sect) __a =	\
				get_unaligned(&p->start_sect);	\
				__le32_to_cpu(__a); \
			})

/* These three have identical behaviour; use the second one if DOS fdisk gets
   confused about extended/logical partitions starting past cylinder 1023. */
#define DOS_EXTENDED_PARTITION 5
#define LINUX_EXTENDED_PARTITION 0x85
#define WIN98_EXTENDED_PARTITION 0x0f
#define LINUX_SWAP_PARTITION	0x82
	
struct partition {
	__u8 boot_ind;		/* 0x80 - active */
	__u8 head;		/* starting head */
	__u8 sector;		/* starting sector */
	__u8 cyl;		/* starting cylinder */
	__u8 sys_ind;		/* What partition type */
	__u8 end_head;		/* end head */
	__u8 end_sector;	/* end sector */
	__u8 end_cyl;		/* end cylinder */
	__u32 start_sect;	/* starting sector counting from 0 */
	__u32 nr_sects;		/* nr of sectors in partition */
};

static int current_minor;

static inline int is_extended_partition(struct partition *p)
{
	return (SYS_IND(p) == DOS_EXTENDED_PARTITION ||
		SYS_IND(p) == WIN98_EXTENDED_PARTITION ||
		SYS_IND(p) == LINUX_EXTENDED_PARTITION);
}

/*
 * Create devices for each logical partition in an extended partition.
 * The logical partitions form a linked list, with each entry being
 * a partition table with two entries.  The first entry
 * is the real data partition (with a start relative to the partition
 * table start).  The second is a pointer to the next logical partition
 * (with a start relative to the entire extended partition).
 * We do not create a Linux partition for the partition tables, but
 * only for the actual data partitions.
 */

#define MSDOS_LABEL_MAGIC		0xAA55

static void extended_partition(char *device, int fd, unsigned long startsec,
			       unsigned long size)
{
    struct partition *p;
    unsigned int first_sector, first_size, this_sector, this_size;
    unsigned char data[512];

    first_sector = startsec;
    first_size = size;
    this_sector = first_sector;
    this_size = size;
    
    while (1) {
	if (!sread( fd, this_sector, data ))
	    return;
	if ((*(unsigned short *) (data+510)) != __cpu_to_le16(MSDOS_LABEL_MAGIC))
	    return;
	
	p = (struct partition *) (0x1BE + data);
	
	/*
	 * Usually, the first entry is the real data partition,
	 * the 2nd entry is the next extended partition, or empty,
	 * and the 3rd and 4th entries are unused.
	 * However, DRDOS sometimes has the extended partition as
	 * the first entry (when the data partition is empty),
	 * and OS/2 seems to use all four entries.
	 */
	
	/* 
	 * First process the data partition(s)
	 */
	fdisk_add_partition(device, current_minor++, SYS_IND(p),
				NR_SECTS(p)/2);
	p++;

	/*
	 * Next, process the (first) extended partition, if present.
	 * (So far, there seems to be no reason to make
	 *  extended_partition()  recursive and allow a tree
	 *  of extended partitions.)
	 * It should be a link to the next logical partition.
	 * Create a minor for this just long enough to get the next
	 * partition table.  The minor will be reused for the next
	 * data partition.
	 */
	if(! (NR_SECTS(p) && is_extended_partition(p)))
	    return;  /* nothing left to do */
	
	this_sector = first_sector + START_SECT(p);
	this_size = NR_SECTS(p);
    }
}

int parse_msdos_partition(char *device, int fd)
{
    int i, minor = 1;
    struct partition *p;
    unsigned char data[512];

    if (!sread( fd, 0, data ))
	return -1;
    if (*(unsigned short *) (0x1fe + data) != __cpu_to_le16(MSDOS_LABEL_MAGIC))
	return 0;
    p = (struct partition *) (0x1be + data);

    current_minor = minor+4; /* first "extra" minor (for extended partitions) */
    for (i=1 ; i<=4 ; minor++,i++,p++) {
	if (!NR_SECTS(p) || !SYS_IND(p))
	    continue;
	fdisk_add_partition(device, minor, SYS_IND(p), NR_SECTS(p)/2);
	if (is_extended_partition(p)) {
	    extended_partition(device, fd, START_SECT(p), NR_SECTS(p));
	}
    }
    return 1;
}
